# -*- coding: utf-8 -*-


def is_floatable(num: any) -> bool:
    """
    https://www.programiz.com/python-programming/examples/check-string-number
    """
    if num == None:
        return None

    else:
        try:
            float(num)
            return True
        except ValueError:
            return False


def convert_str_to_float_single(data_json: list) -> list:
    """
    https://stackoverflow.com/questions/21193682/convert-a-string-key-to-int-in-a-dictionary
            Return and rtype:
            Catatan:
    """

    if isinstance(data_json, dict):
        return [
            {k: float(eval(v)) if is_floatable(v) else v for k, v in data_json.items()}
        ]
    if isinstance(data_json, list):
        data_json = data_json[0]
        return [{k: float(v) if is_floatable(v) else v for k, v in data_json.items()}]


def convert_str_to_float(data_json: list) -> object:
    """
    https://stackoverflow.com/questions/21193682/convert-a-string-key-to-int-in-a-dictionary
            Return and rtype:
            Catatan:
    """
    excluded = [
        "id",
        "reduceOnly",
        "postOnly",
        "liquidation",
        "ioc",
        "market",
        "future",
        "type",
        "side",
        "orderId",
        "time",
        "tradeId",
        "feeCurrency",
        "liquidity",
        "baseCurrency",
        "quoteCurrency",
        "coin",
    ]

    if isinstance(data_json, dict):
        data_json = [data_json]

    if len(data_json) == 1:
        # data_json=data_json[0]
        return convert_str_to_float_single(data_json)

    if len(data_json) > 1:
        list_combination = []
        for i in data_json:
            if isinstance(i, dict):
                data_json = [data_json]
                position = {
                    k: float(v) if (is_floatable(v) and (k) not in excluded) else v
                    for k, v in i.items()
                }
            if isinstance(i, list):
                position = {
                    k: float(v) if (is_floatable(v) and (k) not in excluded) else v
                    for k, v in i.items()
                }
            list_combination.append(position)

        return list_combination


def presisi_pembulatan(angka):
    """
    # Menghitung berapa angka di belakang koma untuk keperluan pembulatan

        Args:
            param1 (float): yang ingin dibulatkan (float: 0.xx atau 1e4)

        Return and rtype:
            berapa angka di belakang koma --> int

        References:
            https://stackoverflow.com/questions/38847690/convert-float-to-string-in-positional-format-without-scientific-notation-and-fa
            https://stackoverflow.com/questions/3886402/how-to-get-numbers-after-decimal-point
            https://stackoverflow.com/questions/26231755/count-decimal-places-in-a-float
    """

    from numpy import format_float_positional as format_float

    angka = str(format_float(angka))

    return int(angka[::-1].find("."))


def rounding(variable, TICK, presisi):
    """
    # Pembulatan berdasarkan tick per instrumen agar sesuai dengan standar harga exchange

        Args:
            param1 (float): angka yang akan dibulatkan
            param2 (float): tick angka
            param3 (int): presisi hasil pembulatan

        Return and rtype:
            hasil pembulatan --> float

    """
    rounding = int(variable / TICK)
    rounding = float(round(rounding * TICK, presisi))

    return rounding


def net_position(selected_transactions: list) -> float:
    """
    net sell: negative, net buy: positive

    """

    if selected_transactions != []:
        try:
            sum_sell = sum(
                [o["amount"] for o in selected_transactions if o["direction"] == "sell"]
            )  # sell = + sign
            sum_buy = sum(
                [o["amount"] for o in selected_transactions if o["direction"] == "buy"]
            )

            #! -1 + 1 = 0, -10+10 = 0, [] = 0, None = 0. [] = No transcations, diff with net = 0 (could affect to leverage)
            #! solution = made another controls for leverage/modify output/preventive
            return sum_buy - sum_sell

        except:
            return ([o["size"] for o in selected_transactions])[0]  # sell = (-) sign

    else:
        return selected_transactions


def get_nearest_tick(price: float, tick: float) -> float:
    """ """
    len_tick = len(str(tick)) - 2

    return round((int(price / tick)) * tick, len_tick)


def get_closest_value(lst, k) -> float:
    """

    References:
        https://www.geeksforgeeks.org/python-find-closest-number-to-k-in-given-list/
    """

    import bisect

    # lst = [3.64, 5.2, 9.42, 9.35, 8.5, 8]
    # k = 9.1

    lst.sort()
    index = bisect.bisect_left(lst, k)
    if index == 0:
        closest = lst[0]
    elif index == len(lst):
        closest = lst[-1]
    else:
        before = lst[index - 1]
        after = lst[index]
        closest = before if after - k > k - before else after
    # print(closest)

    return closest



# -*- coding: utf-8 -*-
import base64


def get_oci_key(keyOCID) -> list:
    import oci

    # Create vaultsclient using the default config file (\.oci\config) for auth to the API
    config = oci.config.from_file()
    vaultclient = oci.vault.VaultsClient(config)

    # Get the secret
    secretclient = oci.secrets.SecretsClient(config)
    secretcontents = secretclient.get_secret_bundle(secret_id=keyOCID)

    # Decode the secret from base64 and print
    keybase64 = secretcontents.data.secret_bundle_content.content
    keybase64bytes = keybase64.encode("ascii")
    keybytes = base64.b64decode(keybase64bytes)
    key = keybytes.decode("ascii")

    return key



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# built ins
import asyncio
import telegram
import os
import sys
from asyncio import run, gather

# -----------------------------------------------------------------------------

this_folder = os.path.dirname(os.path.abspath(__file__))
root_folder = os.path.dirname(os.path.dirname(this_folder))
sys.path.append(root_folder + "/python")
sys.path.append(this_folder)

# -----------------------------------------------------------------------------

import ccxt.async_support as ccxt  # noqa: E402

# -----------------------------------------------------------------------------

# installed
from loguru import logger as log

# user defined formula
from configuration import config


from messaging import (
    get_published_messages,
    telegram_bot as tlgrm,
)
from utilities import (
    string_modification as str_mod,
    system_tools,
    time_modification as time_mod,
)


async def relaying_result(
    client_redis: object,
    config_app: list,
    redis_channels: list,
) -> None:
    """ """

    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        abnormal_trading_notices_channel: str = redis_channels[
            "abnormal_trading_notices"
        ]

        # prepare channels placeholders
        channels = [
            abnormal_trading_notices_channel,
        ]

        # subscribe to channels
        [await pubsub.subscribe(o) for o in channels]

        tlgrm_id = config.main_dotenv("telegram-binance")
        TOKEN = tlgrm_id["bot_token"]
        chat_id = tlgrm_id["bot_chatid"]

        bot = telegram.Bot(token=TOKEN)

        exchange = ccxt.binance()

        result = []

        ONE_SECOND = 1000

        one_minute = ONE_SECOND * 60

        one_hour = one_minute * 60

        fifteen_min = 15 * one_minute

        send_tlgrm = False

        while True:

            try:

                message_byte = await pubsub.get_message()

                params = await get_published_messages.get_redis_message(message_byte)

                data = params["data"]

                message_channel = params["channel"]

                if abnormal_trading_notices_channel in message_channel:

                    noticeType = data["noticeType"]
                    symbol = data["symbol"]
                    eventType = data["eventType"]
                    period = data["period"]
                    priceChange = data["priceChange"]

                    if "MINUTE" in period:

                        log.info(data)

                        tf_int = str_mod.extract_integers_from_text(period)
                        timeframe = f"{tf_int}m"
                        limit = 9

                        if "VOLUME" in noticeType:

                            if "HIGH" in eventType:

                                datetime = exchange.iso8601(exchange.milliseconds())

                                main = (
                                    f"{symbol} experienced HIGHER volume than average\n"
                                )
                                extra_info = (
                                    f"TF: {timeframe}, price change: {priceChange}\n"
                                )
                                wording = f"{main} {extra_info} {datetime}"

                                await bot.send_message(
                                    text=wording,
                                    chat_id=chat_id,
                                )

                        else:

                            is_fluctuated = await compute_price_changes_result(
                                exchange,
                                symbol,
                                timeframe,
                                limit,
                            )

                            if is_fluctuated["wording"]:

                                current_timestamp = time_mod.get_now_unix_time()

                                if result:

                                    symbol_is_exist = [
                                        o
                                        for o in result
                                        if o["symbol"] == is_fluctuated["symbol"]
                                    ]

                                    if symbol_is_exist:

                                        max_timestamp = max(
                                            [o["timestamp"] for o in symbol_is_exist]
                                        )

                                        symbol_with_max_timestamp = [
                                            o
                                            for o in symbol_is_exist
                                            if o["timestamp"] == max_timestamp
                                        ]

                                        timestamp_expired = is_timestamp_expired(
                                            current_timestamp,
                                            max_timestamp,
                                            fifteen_min,
                                        )

                                        if timestamp_expired:
                                            result.remove(symbol_with_max_timestamp[0])

                                            updating_cache(
                                                result,
                                                is_fluctuated["symbol"],
                                                current_timestamp,
                                            )

                                            send_tlgrm = True

                                    else:

                                        updating_cache(
                                            result,
                                            is_fluctuated["symbol"],
                                            current_timestamp,
                                        )

                                        send_tlgrm = True

                                else:

                                    updating_cache(
                                        result,
                                        is_fluctuated["symbol"],
                                        current_timestamp,
                                    )

                                    send_tlgrm = True

                            if send_tlgrm:

                                await bot.send_message(
                                    text=is_fluctuated["wording"],
                                    chat_id=chat_id,
                                )

                                send_tlgrm = False

            except Exception as error:
                
                log.debug(error)

                if "binance does not have market symbol" in error:
                    pass

                else:

                    await tlgrm.telegram_bot_sendtext(
                        f"relaying_result - {error}",
                        "general_error",
                    )

                    system_tools.parse_error_message(error)

                continue

            finally:
                await asyncio.sleep(0.001)

    except Exception as error:
        
        log.warning(error)
        
        if "binance does not have market symbol" in error:
            pass

        else:

            system_tools.parse_error_message(error)

            await tlgrm.telegram_bot_sendtext(
                f"relaying_result - {error}",
                "general_error",
            )


async def sending_telegram(data: list) -> None:

    """
    noticeType = [
        PRICE_BREAKTHROUGH,
        PRICE_CHANGE,
        PRICE_FLUCTUATION,
        VOLUME_PRICE
        ]

    eventType = [
        DOWN_BREAKTHROUGH,
        RISE_AGAIN,
        DROP_BACK,
        UP_BREAKTHROUGH,
        UP_2,
        UP_1,
        DOWN_1,
        DOWN_2,
        HIGH_VOLUME_RISE_3,
        HIGH_VOLUME_RISE_2,
        HIGH_VOLUME_RISE_1,
        ]

    period = [
        WEEK_1,
        DAY_1,
        MONTH_1,
        MINUTE_5,
        HOUR_2
        ]

    example:
        {
            'type': 'VOLUME_PRICE',
            'symbol': 'LISTAUSDT',
            'event': 'HIGH_VOLUME_RISE_1',
            'price change': 0.10529986,
            'period': 'MINUTE_15'
            }


    """

    pass


async def compute_price_changes_result(
    exchange: str,
    symbol: str,
    timeframe: str,
    limit: int,
) -> str:

    """ """

    wording = ""

    ohlcv = await get_ohlcv(exchange, symbol, timeframe, limit)

    ticker = await get_ticker(exchange, symbol)

    if len(ohlcv):
        last_candle = ohlcv[limit - 1]
        datetime = ticker["datetime"]
        last = ticker["last"]
        open = last_candle[1]
        close = last_candle[3]

        delta_close = close - open
        delta_close_pct = abs(((delta_close / open)))

        delta_current = last - open
        delta_current_pct = abs(round((delta_current / open), 2))

        THRESHOLD = 3 / 100

        if delta_close_pct >= THRESHOLD:

            if delta_close > 0:
                move = "HIGHER"
            if delta_close < 0:
                move = "LOWER"

            main = f"{symbol} closing is {round(delta_close_pct*100,2)}%  {move} than its opening \n"
            extra_info = (
                f"TF: {timeframe}, Open: {open}, Close: {close}, Current: {last}\n"
            )
            wording = f"{main} {extra_info} {datetime}"

        if delta_current_pct >= THRESHOLD:

            if delta_current > 0:
                move = "HIGHER"
            if delta_current < 0:
                move = "LOWER"

            main = f"{symbol} current price is {round(delta_current_pct*100,2)}%  {move} than its opening \n"
            extra_info = f"TF: {timeframe}, Open: {open}, Current: {last}\n"
            wording = f"{main} {extra_info} {datetime}"

    await exchange.close()

    return dict(
        wording=wording,
        symbol=symbol,
    )


async def get_ohlcv(
    exchange: str,
    symbol: str,
    timeframe: str,
    limit: int,
    since: int = None,
) -> dict:

    return await exchange.fetch_ohlcv(
        symbol,
        timeframe,
        since,
        limit,
    )


def is_timestamp_expired(
    current_timestamp: int,
    symbol_timestamp: int,
    threshold: int,
) -> dict:

    """
    check if the timestamp is expired
    """
    return (current_timestamp - symbol_timestamp) > threshold


def updating_cache(
    result: list,
    symbol: str,
    current_timestamp: int,
) -> dict:

    """ """
    res = {}

    if symbol:

        res.update({"symbol": symbol})
        res.update({"timestamp": current_timestamp})

        result.append(res)


async def get_ticker(
    exchange: str,
    symbol: str,
) -> dict:

    """
    example: {
        'symbol': 'HARD/USDT',
        'timestamp': 1744501380384,
        'datetime': '2025-04-12T23:43:00.384Z',
        'high': 0.0319,
        'low': 0.0202,
        'bid': 0.0224,
        'bidVolume': 2176.0,
        'ask': 0.0226,
        'askVolume': 4170.0,
        'vwap': 0.02514621,
        'open': 0.0319,
        'close': 0.0226,
        'last': 0.0226,
        'previousClose': 0.0318,
        'change': -0.0093,
        'percentage': -29.154,
        'average': 0.0272,
        'baseVolume': 25033279.0,
        'quoteVolume': 629492.088,
        'markPrice': None,
        'indexPrice': None,
        'info': {
            'symbol': 'HARDUSDT',
            'priceChange': '-0.00930000',
            'priceChangePercent': '-29.154',
            'weightedAvgPrice': '0.02514621',
            'prevClosePrice': '0.03180000',
            'lastPrice': '0.02260000',
            'lastQty': '2252.00000000',
            'bidPrice': '0.02240000',
            'bidQty': '2176.00000000',
            'askPrice': '0.02260000',
            'askQty': '4170.00000000',
            'openPrice': '0.03190000',
            'highPrice': '0.03190000',
            'lowPrice': '0.02020000',
            'volume': '25033279.00000000',
            'quoteVolume': '629492.08800000',
            'openTime': 1744414980384,
            'closeTime': 1744501380384,
            'firstId': 41880405,
            'lastId': 41897290,
            'count': 16886
            }
            }

    """

    return await exchange.fetch_ticker(symbol)



# built ins
import asyncio

# import orjson
import httpx

# user defined formula
from configuration import config


async def private_connection(
    endpoint: str,
    connection_url: str = "https://api.telegram.org/bot",
) -> None:

    async with httpx.AsyncClient() as session:

        respons = await session.get(connection_url + endpoint)

        return respons.json()


async def telegram_bot_sendtext(
    bot_message: str, purpose: str = "general_error"
) -> str:
    """
    # simple telegram
    #https://stackoverflow.com/questions/32423837/telegram-bot-how-to-get-a-group-chat-id
    """

    try:
        bot_token = config.main_dotenv("telegram-failed_order")["bot_token"]

    except:
        bot_token = config.main_dotenv("telegram-failed_order")["BOT_TOKEN"]

    if purpose == "failed_order":
        try:
            try:
                bot_chatID = config.main_dotenv("telegram-failed_order")[
                    "BOT_CHATID_FAILED_ORDER"
                ]
            except:
                bot_chatID = config.main_dotenv("telegram-failed_order")["bot_chatID"]
        except:
            bot_chatID = config.main_dotenv("telegram-failed_order")["bot_chatid"]

    if purpose == "general_error":
        try:
            try:
                bot_chatID = config.main_dotenv("telegram-general_error")["bot_chatid"]
            except:
                bot_chatID = config.main_dotenv("telegram-general_error")["bot_chatID"]
        except:
            bot_chatID = config.main_dotenv("telegram-general_error")[
                "BOT_CHATID_GENERAL_ERROR"
            ]

    connection_url = "https://api.telegram.org/bot"

    endpoint = (
        bot_token
        + ("/sendMessage?chat_id=")
        + bot_chatID
        + ("&parse_mode=HTML&text=")
        + str(bot_message)
    )

    return await private_connection(endpoint=endpoint, connection_url=connection_url)



# -*- coding: utf-8 -*-

# built ins
import asyncio
import json
import time
from datetime import datetime, timezone

# installed
import orjson
import uvloop
import websockets
from dataclassy import dataclass, fields
from loguru import logger as log

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

# user defined formula
from configuration import config
from messaging.telegram_bot import telegram_bot_sendtext
from utilities import system_tools, string_modification as str_mod


def parse_dotenv(sub_account: str) -> dict:
    return config.main_dotenv(sub_account)


def get_timestamp():
    return int(time.time() * 1000)


@dataclass(unsafe_hash=True, slots=True)
class StreamingDataBinance:
    """
    https://www.binance.com/en/support/faq/detail/18c97e8ab67a4e1b824edd590cae9f16

    """

    sub_account_id: str
    client_id: str = fields
    client_secret: str = fields
    # Async Event Loop
    loop = asyncio.get_event_loop()
    ws_connection_notice_url: str = "wss://bstream.binance.com:9443/stream?"
    # Instance Variables
    websocket_client: websockets.WebSocketClientProtocol = None
    refresh_token: str = None
    refresh_token_expiry_time: int = None

    async def ws_manager(
        self,
        client_redis: object,
        redis_channels,
        queue_general: object,
    ) -> None:

        async with websockets.connect(
            self.ws_connection_notice_url,
            ping_interval=None,
            compression=None,
            close_timeout=60,
        ) as self.websocket_client:

            try:

                # timestamp = get_timestamp()

                # encoding_result = hashing(timestamp, self.client_id, self.client_secret)

                msg = {
                    "method": "SUBSCRIBE",
                    "params": ["abnormaltradingnotices"],
                    "id": 1,
                }

                while True:

                    ws_channel = ["abnormaltradingnotices"]

                    await self.ws_operation(
                        operation="SUBSCRIBE",
                        ws_channel=ws_channel,
                        source="ws",
                    )

                    while True:

                        # Receive WebSocket messages
                        message: bytes = await self.websocket_client.recv()
                        message: dict = orjson.loads(message)

                        if message:
                            data = message.get("data", None)

                            # queing message to dispatcher
                            await queue_general.put(message)

            except Exception as error:

                system_tools.parse_error_message(error)

                await telegram_bot_sendtext(
                    (f"""data producer - {error}"""),
                    "general_error",
                )

    async def establish_heartbeat(self) -> None:
        """
        reference: https://github.com/ElliotP123/crypto-exchange-code-samples/blob/master/deribit/websockets/dbt-ws-authenticated-example.py

        Requests DBT's `public/set_heartbeat` to
        establish a heartbeat connection.
        """
        msg: dict = {
            "jsonrpc": "2.0",
            "id": 9098,
            "method": "public/set_heartbeat",
            "params": {"interval": 10},
        }

        try:
            await self.websocket_client.send(json.dumps(msg))

        except Exception as error:

            system_tools.parse_error_message(error)

            await telegram_bot_sendtext(
                (f"""data producer establish_heartbeat - {error}"""),
                "general_error",
            )

    async def heartbeat_response(self) -> None:
        """
        Sends the required WebSocket response to
        the Deribit API Heartbeat message.
        """
        msg: dict = {
            "jsonrpc": "2.0",
            "id": 8212,
            "method": "public/test",
            "params": {},
        }

        try:

            await self.websocket_client.send(json.dumps(msg))

        except Exception as error:

            system_tools.parse_error_message(error)

            await telegram_bot_sendtext(
                (f"""data producer heartbeat_response - {error}"""),
                "general_error",
            )

    async def ws_auth(self) -> None:
        """
        Requests DBT's `public/auth` to
        authenticate the WebSocket Connection.
        """
        msg: dict = {
            "jsonrpc": "2.0",
            "id": 9929,
            "method": "public/auth",
            "params": {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            },
        }

        try:
            await self.websocket_client.send(json.dumps(msg))

        except Exception as error:

            system_tools.parse_error_message(error)

            await telegram_bot_sendtext(
                (f"""data producer - {error}"""),
                "general_error",
            )

    async def ws_refresh_auth(self) -> None:
        """
        Requests DBT's `public/auth` to refresh
        the WebSocket Connection's authentication.
        """
        while True:

            now_utc = datetime.now(timezone.utc)

            if self.refresh_token_expiry_time is not None:

                if now_utc > self.refresh_token_expiry_time:

                    msg: dict = {
                        "jsonrpc": "2.0",
                        "id": 9929,
                        "method": "public/auth",
                        "params": {
                            "grant_type": "refresh_token",
                            "refresh_token": self.refresh_token,
                        },
                    }

                    await self.websocket_client.send(json.dumps(msg))

            await asyncio.sleep(150)

    async def ws_operation(
        self,
        operation: str,
        ws_channel: str,
        source: str = "ws",
    ) -> None:
        """ """
        sleep_time: int = 0.05

        await asyncio.sleep(sleep_time)

        id = 1

        msg: dict = {}

        if "ws" in source:
            extra_params: dict = dict(
                id=id,
                method=f"{operation}",
                params=ws_channel,
            )

            msg.update(extra_params)

            await self.websocket_client.send(json.dumps(msg))



# -*- coding: utf-8 -*-

from configparser import ConfigParser

from utilities import system_tools


def catch_error(error, idle: int = None) -> list:
    """ """
    from utilities import system_tools

    system_tools.catch_error_message(error, idle)


class Read_Configuration:
    """
    # Read .env file
    """

    def __init__(self):
        self.params = None
        self.conn = None

    def config(self, filename: str, section: str) -> dict:
        # create parser
        parser = ConfigParser()

        # read file config
        parser.read(filename)

        # prepare place holder for file config read result
        parameters = {}

        if self.params is None:
            # check file section
            if parser.has_section(section):
                params = parser.items(section)
                for param in params:
                    parameters[param[0]] = param[1]

            # if section is not provided
            else:
                raise Exception(
                    "Section {0} not found in the {1} file".format(section, filename)
                )

        return parameters


def main_dotenv(header: str = "None", filename: str = ".env") -> dict:
    """
    https://www.python-engineer.com/posts/run-python-github-actions/
    """

    # Initialize credentials to None
    credentials = None

    try:
        # Set the filename
        # filename = ".env"
        config_path = system_tools.provide_path_for_file(filename)

        # Create a Read_Configuration object
        Connection = Read_Configuration()

        credentials = Connection.config(config_path, header)

    # to accomodate transition phase. Will be deleted
    except:
        import os
        from os.path import dirname, join

        from dotenv import load_dotenv

        dotenv_path = join(dirname(__file__), ".env")
        load_dotenv(dotenv_path)

        # github env
        credentials = os.environ
        # log.info (credentials)

    return credentials


if __name__ == "__main__":
    try:
        test = main_dotenv("telegram-failed_order")
        print(test)

        test = main_dotenv("deribit-147691")
        print(test)

    except KeyboardInterrupt:
        catch_error(KeyboardInterrupt)

    except Exception as error:
        catch_error(error, 30)
